defstar {
 	name { ControlFIFO}
	domain { DE }
	author { Dorgham Sisalem }
        copyright { SEE $PTOLEMY_SIMULATIONS/TCP/copyright }

        desc   { This star simulates the buffer of the application.
	         }

	input { name { demand}
                type { float }
	        desc { the application has finished consuming the last packet}
	}
	input { name { inData }
                type { anytype}
	        desc { new data}
	}

	output {
		name {outData}
		type {=inData}
	        desc { send data to the application}
	}
	output {
		name {size}
		type {int}
		desc { Size of Queue. Triggered by demand inputs. }
	}
	output {
		name {overflow}
		type {=inData}
		desc {
	        Arrival data that cannot be queued due to capacity limit.
		}
	}

	protected {
		int infinite;
	}
	defstate {
		name {capacity}
		type {int}
		default {"-1"}
		desc { Maximum size of the queue. If <0, capacity is infinite. }
	}

	constructor {
		// Indicate to the scheduler that size outputs are triggered
		// demand inputs.
		// simultaneous inData events should be available in the
		// same firing with demand events
         	   demand.before(inData);
	}
	setup {
		infinite = (int(capacity) < 0);
		control=1;
		First=1;
		zapQueue();
	}

	ccinclude {
		"DataStruct.h"
	}
	protected {
		Queue queue;
	}
	method {
		name { enqueue }
		access { protected }
		arglist { "()" }
		type { void }
		code {
		    if (infinite || Qsize() < int(capacity)) {
			queue.put(inData.get().clone());
		    } else {
			// Data is lost
			overflow.put(completionTime) = inData.get();
		    }
		}
	}
	method {
		name { dequeue }
		access { protected }
		arglist { "()" }
		type { Pointer }
		code {
		    return queue.get();
		}
	}
	method {
		name { handleOverflow }
		access { protected }
		arglist { "()" }
		type { void }
		code {
		    // Nothing to do
		}
	}
	method {
		name { Qsize }
		access { protected }
		arglist { "()" }
		type { int }
		code {
		    return queue.length();
		}
	}
	method {
		name { zapQueue }
		code {
			while (Qsize() > 0) {
				Particle* pp = (Particle*) queue.get();
				pp->die();
			}
			queue.initialize();
		      }
               }
	protected{
	  int control;
	  int First;
	}
	go {
	    completionTime = arrivalTime;

//****************************************************************************
// The application finished consuming a packet an can take another one
//****************************************************************************

	    if (demand.dataNew)
	    {
	      demand.dataNew= FALSE;
	      size.put(completionTime)<<Qsize(); // update the window size
       	      if(Qsize())
	      {
	
		  Particle* pp = (Particle*) dequeue();
		  outData.put(completionTime) = *pp;
		  pp->die();
		  control=0; // application is busy
	      }
	      else
	          control=1;
	    }

//*************************************************************************
// If there is new inData, store it if the application is busy
//*************************************************************************

	    if (inData.dataNew)
	    {
	      inData.dataNew=FALSE;
	      if(!Qsize()&&control)
	      {
	        if(First>0)
	        {
	           First=0;
		   size.put(completionTime)<<1;
		}
		outData.put(completionTime) =inData%0;
	        control=0;
	      }
	      else
	      {
		enqueue();
	        //size.put(completionTime)<<Qsize();
	      }
	    }

	}

        destructor {
		zapQueue();
	  }
}
