 defstar {
 	name {EPRCA_Switch }
	domain { DE }
	author { Dorgham Sisalem}
        copyright { SEE $PTOLEMY_SIMULATIONS/copyright }

	desc {
	}

	input { name { IN_RM}
                type { message}
	        desc { incoming RM Cells}
	}

	input { name { RET_RM}
                type { message}
	        desc { Returning RM Cells}
	}

	input { name { Congestion}
                type { float}
	        desc { indicate the congestion state}
	}

	output { name { OUT_RM}
                 type { message}
	         desc { Send RM Cells to the Destination}
        }

	output { name { NEW_RM}
                 type { message}
	         desc { updated RM Cells}
        }

	defstate {
		name { PCR}
		type { float}
		default {0 }
		desc { Peak Cell Rate}
	}

	defstate {
		name { IMR}
		type { float}
		default {0 }
		desc { Initial rate for MACR}
	}

	defstate {
		name { VCS}
		type { float}
		default { 0.875}
		desc { VC seperator}
	}

	defstate {
		name { AV}
		type { float}
		default { 0.0625}
		desc { Exponential Averaging Factor}
	}

	defstate {
		name { DQT}
		type { int}
		default { 100}
		desc { High Queue Limit}
	}

	defstate {
		name { MRF}
		type { float}
		default { 0.25}
		desc { Major Reduction Factor}
	}

	defstate {
		name { DPF}
		type { float}
		default { 0.875}
		desc { Down Pressure Factor}
	}

	defstate {
		name { ERF}
		type { float}
		default { 0.9375}
		desc { Explicit Reduction Factor}
	}

	defstate {
		name { mode}
		type { int}
		default {0 }
		desc { 0=Explicit,1=Binary}
	}

	hinclude { "ATMCell.h" ,<math.h>,<minmax.h>}

	defstate {
		name { fileName }
		type { string }
		default { "<stdout>" }
		desc { Filename for output }
	}
	hinclude { "pt_fstream.h" }
	protected {
		pt_ofstream *p_out;
	}

	protected { 
			int congested;
			float ACR,MACR;
		  }

//*****************************************************************************/
// initialization: 
//*****************************************************************************/

	setup { 
		IMR=PCR/100;
		MACR=IMR;
		congested=0;
              }

//*****************************************************************************/
//*****************************************************************************/

	begin  {  
                LOG_DEL; delete p_out;
		LOG_NEW; p_out = new pt_ofstream(fileName);
		}

//****************************************************************************/
// set scheduling priorities
//****************************************************************************/

	constructor { 
                      p_out = 0;
               	    }
        wrapup {
	        LOG_DEL; delete p_out;
		p_out = 0;
		}
	destructor { LOG_DEL; delete p_out;}

	go {
                pt_ofstream& OUTPUT = *p_out;
                completionTime= arrivalTime;

		if(Congestion.dataNew)
		{
			Congestion.dataNew=FALSE;
			float cong=Congestion%0;
			congested= (int) cong;
		} //end Congestion

		if(IN_RM.dataNew)
		{
			Envelope Envp;
	                IN_RM.get().getMessage(Envp);
       	            	ATMCell* cellPtr=(ATMCell *)
                    	  	Envp.myData();
                  	if(cellPtr->isOFF(28))
			{
				OUT_RM.put(completionTime)<<Envp;
				return;
			}
			float ACR=cellPtr->readACR();
			float M_V=(float)MACR*(float)VCS;
OUTPUT<<"IN_ER "<<cellPtr->readER()<<" M_V "<<M_V<<" ACR "<<ACR<<" CON "<<congested<<"\n";
			if(((congested>0)&&(ACR<MACR))||
				((congested==0)&&(ACR>M_V)))
			{
				float A_M=ACR-MACR;
//OUTPUT<<" A_M "<<A_M;
				float A_M_V=(float)A_M*(float)AV;
				MACR=MACR+A_M_V;
OUTPUT<< " MACR "<<MACR<<" at "<<completionTime<<" \n";
			}
			OUT_RM.put(completionTime)<<Envp;
//OUTPUT<<" SEN RM at "<<completionTime<<" \n";
		} //end IN_RM

		if(RET_RM.dataNew)
		{
			Envelope Envp;
	                RET_RM.get().getMessage(Envp);
       	            	ATMCell* cellPtr=(ATMCell *)
                    	  	Envp.myData();
			float ACR=cellPtr->readACR();
                        LOG_NEW;
                        ATMCell* RM_Cell = new ATMCell(0);
                        RM_Cell->set(28);
                        RM_Cell->setSource(cellPtr->readSource());
                        RM_Cell->setVC(cellPtr->readVC());
                        RM_Cell->setDestination(cellPtr->readDestination());
                        RM_Cell->setDIR(0);
                        RM_Cell->setACR(ACR);
                        RM_Cell->setCI(cellPtr->readCI());
                        RM_Cell->setER(cellPtr->readER());
OUTPUT<<" cong "<<congested<<" DEST "<<RM_Cell->readDestination()<<" ER "<<cellPtr->readER()<<" at "<<completionTime<<"\n";
			if(congested)
			{
				if(congested>DQT)
				{
					if(mode==1)
					{
                           			RM_Cell->setCI(1);
					}
					if(mode==0)	
					{
				             float cell_ER=cellPtr->readER();
					     float ME = (float)MACR*(float)MRF;
						float ER=min(ME,cell_ER);
OUTPUT<<" MACR1 "<<MACR<<" ER "<<ER<<" ME "<<ME<<" cell_ER "<<cell_ER<<"\n";
                        		     RM_Cell->setER(ER);
					}	
				}
				else
				{
					float M_D=(float) MACR * (float) DPF;
OUTPUT<<" ACR "<<ACR<<" M_D "<<M_D<<" MACR "<<MACR<<" DPF "<<DPF<<"\n";
					if(ACR>M_D)
					{
				     	     if(mode==1)
					     {
                           			RM_Cell->setCI(1);
					     }
					     if(mode==0)	
					     {
						float ME=(float) MACR * (float) ERF;
						float cell_ER=cellPtr->readER();
						float ER=min(ME,cell_ER);
OUTPUT<<" MACR "<<MACR<<" ER "<<ER<<" ME "<<ME<<" cell_ER "<<cell_ER<<"\n";
                        			RM_Cell->setER(ER);
					     }
					}
				}	
			}
                        Envelope envp( *RM_Cell);
			NEW_RM.put(completionTime)<<envp;
		}// end RET_RM
	   } // end go
} // end defstar
