static const char file_id[] = "DEERAlgorithm.pl";
// .cc file generated from DEERAlgorithm.pl by ptlang
/*
 * copyright (c) SEE $PTOLEMY_SIMULATIONS/copyright
 */

#ifdef __GNUG__
#pragma implementation
#endif

#include "DEERAlgorithm.h"

const char *star_nm_DEERAlgorithm = "DEERAlgorithm";

const char* DEERAlgorithm :: className() const {return star_nm_DEERAlgorithm;}

ISA_FUNC(DEERAlgorithm,DEStar);

Block* DEERAlgorithm :: makeNew() const { LOG_NEW; return new DEERAlgorithm;}

DEERAlgorithm::DEERAlgorithm ()
{
	setDescriptor("");
	addPort(IN_RM.setPort("IN_RM",this,MESSAGE));
	addPort(RET_RM.setPort("RET_RM",this,MESSAGE));
	addPort(Congestion.setPort("Congestion",this,INT));
	addPort(OUT_RM.setPort("OUT_RM",this,MESSAGE));
	addPort(NEW_RM.setPort("NEW_RM",this,MESSAGE));
	addState(PCR.setState("PCR",this,"0","Peak Cell Rate"));
	addState(IMR.setState("IMR",this,"0","Initial rate for MACR"));
	addState(VCS.setState("VCS",this,"0.875","VC seperator"));
	addState(AV.setState("AV",this,"0.0625","Exponential Averaging Factor"));
	addState(DQT.setState("DQT",this,"100","High Queue Limit"));
	addState(MRF.setState("MRF",this,"0.25","Major Reduction Factor"));
	addState(DPF.setState("DPF",this,"0.875","Down Pressure Factor"));
	addState(ERF.setState("ERF",this,"0.9375","Explicit Reduction Factor"));
	addState(mode.setState("mode",this,"0","0=Explicit,1=Binary"));
	addState(TLow.setState("TLow",this,"0","TLow"));
	addState(RM_Proc.setState("RM_Proc",this,"","RM_Proc"));
	addState(debug.setState("debug",this,"0","debug"));
	addState(fileName.setState("fileName",this,"<stdout>","Filename for output"));

# line 153 "DEERAlgorithm.pl"
p_out = 0;
}

void DEERAlgorithm::wrapup() {
# line 156 "DEERAlgorithm.pl"
LOG_DEL; delete p_out;
		p_out = 0;
}

void DEERAlgorithm::begin() {
# line 144 "DEERAlgorithm.pl"
LOG_DEL; delete p_out;
		LOG_NEW; p_out = new pt_ofstream(fileName);
}

DEERAlgorithm::~DEERAlgorithm() {
# line 159 "DEERAlgorithm.pl"
LOG_DEL; delete p_out;
}

void DEERAlgorithm::setup() {
# line 135 "DEERAlgorithm.pl"
IMR=PCR/100;
		MACR=IMR;
		congested=0;
}

void DEERAlgorithm::go() {
# line 162 "DEERAlgorithm.pl"
pt_ofstream& OUTPUT = *p_out;
                completionTime= arrivalTime;

		InDEMPHIter inIter(RET_RM);
		InDEPort* inPtr;
		if(Congestion.dataNew)
		{
			Congestion.dataNew=FALSE;
			float cong=Congestion%0;
			congested= (int) cong;
		} //end Congestion

		if(IN_RM.dataNew)
	        {
			Envelope Envp;
	                IN_RM.get().getMessage(Envp);
       	            	ATMCell* cellPtr=(ATMCell *)
                    	  	Envp.myData();
                  	if(cellPtr->readRM()==0)
			{
				OUT_RM.put(completionTime)<<Envp;
				return;
			}
			float ACR=cellPtr->readACR();
			int VP=cellPtr->readDestination();
			int VC=cellPtr->readVC();
			float M_V=(float)MACR*(float)VCS;
if(debug)
OUTPUT<<"IN_ACR "<<cellPtr->readACR()<<" CON "<<congested<<" VP "<<VP<<" VC "<<VC<<"\n";
			if(((congested>0)&&(ACR<MACR))||
				((congested==0)&&(ACR>M_V)))
			{
				float A_M=ACR-MACR;
				float A_M_V=(float)A_M*(float)AV;
				MACR=MACR+A_M_V;
if(debug)
OUTPUT<< " MACR "<<MACR<<" at "<<completionTime<<" \n";
			}
			OUT_RM.put(completionTime)<<Envp;
//OUTPUT<<" SEN RM at "<<completionTime<<" \n";
		} //end IN_RM

//		if(RET_RM.dataNew)
//******************************************************************
		while ((inPtr = inIter++) != 0) 
		{
		     while (inPtr->dataNew) 
		     {
			Envelope Envp;
			inPtr->get().getMessage(Envp);
       	            	ATMCell* cellPtr=(ATMCell *)
                    	  	Envp.myData();
		        int VP=cellPtr->readDestination();
			int VC=cellPtr->readVC();
		       if((VP!=RM_Proc[0])&&(VP!=RM_Proc[1])&&(VP!=RM_Proc[2])&&(VP!=RM_Proc[3]))
		       {	           
                            Envelope envp( *cellPtr);
			    NEW_RM.put(completionTime)<<envp;
		            return;
			}
			float ACR=cellPtr->readACR();
                        LOG_NEW;
                        ATMCell* RM_Cell = new ATMCell(0);
                        RM_Cell->setRM(1);
                        RM_Cell->setSource(cellPtr->readSource());
                        RM_Cell->setVC(cellPtr->readVC());
                        RM_Cell->setDestination(cellPtr->readDestination());
                        RM_Cell->setDIR(0);
                        RM_Cell->setACR(ACR);
                        RM_Cell->setCI(cellPtr->readCI());
                        RM_Cell->setER(cellPtr->readER());
			if(congested)
			{
				if(congested>DQT)
				{
					if(mode==1)
					{
                           			RM_Cell->setCI(1);
if(debug)OUTPUT<<" SET BINARY DQT FOR "<<RM_Cell->readDestination()<<"\n";
					}
					if(mode==0)	
					{
                           	             RM_Cell->setCI(0);
				             float cell_ER=cellPtr->readER();
					     float ME = (float)MACR*(float)MRF;
					     float ER=min(ME,cell_ER);
if(debug)
OUTPUT<<" MACR1 "<<MACR<<" ER "<<ER<<" ME "<<ME<<" cell_ER "<<cell_ER<<"\n";
                        		     RM_Cell->setER(ER);
					}	
				}
				else
				{
					float M_D=(float) MACR * (float) DPF;
if(debug)
OUTPUT<<" ACR "<<ACR<<" M_D "<<M_D<<" MACR "<<MACR<<" DPF "<<DPF<<"\n";
					if(ACR>M_D)
					{
				     	     if(mode==1)
					     {
if(debug)OUTPUT<<" SET BINARY FOR "<<RM_Cell->readDestination()<<"\n";
                           			RM_Cell->setCI(1);
					     }
					     if(mode==0)	
					     {
			//		       float ME=(float)MACR*(float)ERF;
					       float cell_ER=cellPtr->readER();
					       float ER=min(MACR,cell_ER);
if(debug)
OUTPUT<<" MACR "<<MACR<<" ER "<<ER<<" cell_ER "<<cell_ER<<"\n";
                        			RM_Cell->setER(ER);
					     }
					}
				}	
			}
if(debug)
OUTPUT<<" cong "<<congested<<" VP "<<RM_Cell->readDestination()<<" VC "<<RM_Cell->readVC()<<" ER "<<RM_Cell->readER()<<"  ACR "<<RM_Cell->readACR()<<" at "<<completionTime<<"\n";
                        Envelope envp( *RM_Cell);
			NEW_RM.put(completionTime)<<envp;
		    }
		}// end RET_RM
}

// prototype instance for known block list
static DEERAlgorithm proto;
static RegisterBlock registerBlock(proto,"ERAlgorithm");
