 defstar {
 	name {ERAlgorithm }
	domain { DE }
	author { Dorgham Sisalem}
        copyright { SEE $PTOLEMY_SIMULATIONS/copyright }

	desc {
	}

	input { name { IN_RM}
                type { message}
	        desc { incoming RM Cells}
	}

	inmulti { name { RET_RM}
                type { message}
	        desc { Returning RM Cells}
	}

	input { name { Congestion}
                type { int}
	        desc { indicate the congestion state}
	}

	output { name { OUT_RM}
                 type { message}
	         desc { Send RM Cells to the Destination}
        }

	output { name { NEW_RM}
                 type { message}
	         desc { updated RM Cells}
        }

	defstate {
		name { PCR}
		type { float}
		default {0 }
		desc { Peak Cell Rate}
	}

	defstate {
		name { IMR}
		type { float}
		default {0 }
		desc { Initial rate for MACR}
	}

	defstate {
		name { VCS}
		type { float}
		default { 0.875}
		desc { VC seperator}
	}

	defstate {
		name { AV}
		type { float}
		default { 0.0625}
		desc { Exponential Averaging Factor}
	}

	defstate {
		name { DQT}
		type { int}
		default { 100}
		desc { High Queue Limit}
	}

	defstate {
		name { MRF}
		type { float}
		default { 0.25}
		desc { Major Reduction Factor}
	}
 	defstate {
		name { DPF}
		type { float}
		default { 0.875}
		desc { Down Pressure Factor}
	}

	defstate {
		name { ERF}
		type { float}
		default { 0.9375}
		desc { Explicit Reduction Factor}
	}

	defstate {
		name { mode}
		type { int}
		default {0 }
		desc { 0=Explicit,1=Binary}
	}
        defstate {
                name { TLow}
                type { int}
                default {0 }
        }
 
        defstate {
                name { RM_Proc}
                type { intarray}
        }
        defstate {
                name { debug}
                type { int}
                default {0 }
        }
 

	hinclude { "ATMCell.h" ,<math.h>,<minmax.h>}
	defstate {
		name { fileName }
		type { string }
		default { "<stdout>" }
		desc { Filename for output }
	}
	hinclude { "pt_fstream.h" }
	protected {
		pt_ofstream *p_out;
	}

	protected { 
			int congested;
			float ACR,MACR;
		  }

//*****************************************************************************/
// initialization: 
//*****************************************************************************/

	setup { 
		IMR=PCR/100;
		MACR=IMR;
		congested=0;
              }

//*****************************************************************************/
//*****************************************************************************/

	begin  {  
                LOG_DEL; delete p_out;
		LOG_NEW; p_out = new pt_ofstream(fileName);
		}

//****************************************************************************/
// set scheduling priorities
//****************************************************************************/

	constructor { 
                      p_out = 0;
               	    }
        wrapup {
	        LOG_DEL; delete p_out;
		p_out = 0;
		}
	destructor { LOG_DEL; delete p_out;}

	go {
                pt_ofstream& OUTPUT = *p_out;
                completionTime= arrivalTime;

		InDEMPHIter inIter(RET_RM);
		InDEPort* inPtr;
		if(Congestion.dataNew)
		{
			Congestion.dataNew=FALSE;
			float cong=Congestion%0;
			congested= (int) cong;
		} //end Congestion

		if(IN_RM.dataNew)
	        {
			Envelope Envp;
	                IN_RM.get().getMessage(Envp);
       	            	ATMCell* cellPtr=(ATMCell *)
                    	  	Envp.myData();
                  	if(cellPtr->readRM()==0)
			{
				OUT_RM.put(completionTime)<<Envp;
				return;
			}
			float ACR=cellPtr->readACR();
			int VP=cellPtr->readDestination();
			int VC=cellPtr->readVC();
			float M_V=(float)MACR*(float)VCS;
if(debug)
OUTPUT<<"IN_ACR "<<cellPtr->readACR()<<" CON "<<congested<<" VP "<<VP<<" VC "<<VC<<"\n";
			if(((congested>0)&&(ACR<MACR))||
				((congested==0)&&(ACR>M_V)))
			{
				float A_M=ACR-MACR;
				float A_M_V=(float)A_M*(float)AV;
				MACR=MACR+A_M_V;
if(debug)
OUTPUT<< " MACR "<<MACR<<" at "<<completionTime<<" \n";
			}
			OUT_RM.put(completionTime)<<Envp;
//OUTPUT<<" SEN RM at "<<completionTime<<" \n";
		} //end IN_RM

//		if(RET_RM.dataNew)
//******************************************************************
		while ((inPtr = inIter++) != 0) 
		{
		     while (inPtr->dataNew) 
		     {
			Envelope Envp;
			inPtr->get().getMessage(Envp);
       	            	ATMCell* cellPtr=(ATMCell *)
                    	  	Envp.myData();
		        int VP=cellPtr->readDestination();
			int VC=cellPtr->readVC();
		       if((VP!=RM_Proc[0])&&(VP!=RM_Proc[1])&&(VP!=RM_Proc[2])&&(VP!=RM_Proc[3]))
		       {	           
                            Envelope envp( *cellPtr);
			    NEW_RM.put(completionTime)<<envp;
		            return;
			}
			float ACR=cellPtr->readACR();
                        LOG_NEW;
                        ATMCell* RM_Cell = new ATMCell(0);
                        RM_Cell->setRM(1);
                        RM_Cell->setSource(cellPtr->readSource());
                        RM_Cell->setVC(cellPtr->readVC());
                        RM_Cell->setDestination(cellPtr->readDestination());
                        RM_Cell->setDIR(0);
                        RM_Cell->setACR(ACR);
                        RM_Cell->setCI(cellPtr->readCI());
                        RM_Cell->setER(cellPtr->readER());
			if(congested)
			{
				if(congested>DQT)
				{
					if(mode==1)
					{
                           			RM_Cell->setCI(1);
if(debug)OUTPUT<<" SET BINARY DQT FOR "<<RM_Cell->readDestination()<<"\n";
					}
					if(mode==0)	
					{
                           	             RM_Cell->setCI(0);
				             float cell_ER=cellPtr->readER();
					     float ME = (float)MACR*(float)MRF;
					     float ER=min(ME,cell_ER);
if(debug)
OUTPUT<<" MACR1 "<<MACR<<" ER "<<ER<<" ME "<<ME<<" cell_ER "<<cell_ER<<"\n";
                        		     RM_Cell->setER(ER);
					}	
				}
				else
				{
					float M_D=(float) MACR * (float) DPF;
if(debug)
OUTPUT<<" ACR "<<ACR<<" M_D "<<M_D<<" MACR "<<MACR<<" DPF "<<DPF<<"\n";
					if(ACR>M_D)
					{
				     	     if(mode==1)
					     {
if(debug)OUTPUT<<" SET BINARY FOR "<<RM_Cell->readDestination()<<"\n";
                           			RM_Cell->setCI(1);
					     }
					     if(mode==0)	
					     {
			//		       float ME=(float)MACR*(float)ERF;
					       float cell_ER=cellPtr->readER();
					       float ER=min(MACR,cell_ER);
if(debug)
OUTPUT<<" MACR "<<MACR<<" ER "<<ER<<" cell_ER "<<cell_ER<<"\n";
                        			RM_Cell->setER(ER);
					     }
					}
				}	
			}
if(debug)
OUTPUT<<" cong "<<congested<<" VP "<<RM_Cell->readDestination()<<" VC "<<RM_Cell->readVC()<<" ER "<<RM_Cell->readER()<<"  ACR "<<RM_Cell->readACR()<<" at "<<completionTime<<"\n";
                        Envelope envp( *RM_Cell);
			NEW_RM.put(completionTime)<<envp;
		    }
		}// end RET_RM
	   } // end go
} // end defstar
