defstar {
	name {TriggeredFIFOQueue}
	domain {DE}
	desc { This FIFOQueue passes particles to the output only when
          an demand occurs. The QSize Output is triggered by demand inputs. 
	}
        version { %W% %G% }
	author { Soonhoi Ha and E. A. Lee, Bernd Deffner }
	acknowledge { Incorporates design ideas from Philip Bitar. }
	copyright {
	}
	input {
		name {demand}
		type {anytype}
		desc {  }
	}
	input {
		name {inData}
		type {anytype}
		desc { Particles to be queued. }
	}
	output {
		name {outData}
		type {=inData}
	}
	output {
		name {size}
		type {int}
		desc { Size of Queue. Triggered by demand inputs. }
	}
	output {
		name {overflow}
		type {=inData}
		desc {
Arrival data that cannot be queued due to capacity limit.
		}
	}
	protected {
		int infinite;
	}
	defstate {
		name {capacity}
		type {int}
		default {"-1"}
		desc { Maximum size of the queue. If <0, capacity is infinite. }
	}

	constructor {
		// Indicate to the scheduler that size outputs are triggered
		// demand inputs.
		demand.triggers(size);
		inData.triggers(overflow);

		// simultaneous inData events should be available in the
		// same firing with demand events
		inData.before(demand);

                inData.triggers();
		demand.triggers(outData);
	}
	setup {
		infinite = (int(capacity) < 0);
	}

	ccinclude {
		"DataStruct.h"
	}
	protected {
		Queue queue;
	}
	method {
		name { enqueue }
		access { protected }
		arglist { "()" }
		type { void }
		code {
		    if (infinite || Qsize() < int(capacity)) {
			queue.put(inData.get().clone());
		    } else {
			// Data is lost
			overflow.put(completionTime) = inData.get();
		    }
		}
	}
	method {
		name { dequeue }
		access { protected }
		arglist { "()" }
		type { Pointer }
		code {
		    return queue.get();
		}
	}
	method {
		name { handleOverflow }
		access { protected }
		arglist { "()" }
		type { void }
		code {
		    // Nothing to do
		}
	}
	method {
		name { Qsize }
		access { protected }
		arglist { "()" }
		type { int }
		code {
		    return queue.length();
		}
	}
	method {
		name { zapQueue }
		code {
			while (Qsize() > 0) {
				Particle* pp = (Particle*) queue.get();
				pp->die();
			}
			queue.initialize();
		      }
                }

	setup {
		zapQueue();
	}

	go {
	    completionTime = arrivalTime;
	    if (demand.dataNew) {
	      demand.dataNew= FALSE;
              if (Qsize() > 0)
                {
		  Particle* pp = (Particle*) dequeue();
		  outData.put(completionTime) = *pp;
		  pp->die();
		}	    
	      // output the queue size
	      //size.put(completionTime) << Qsize();

	    }

	    // If there is new inData, store it
	    while (inData.dataNew) {
		enqueue();
	        size.put(completionTime) << Qsize();
		inData.getSimulEvent();
	    }

	    // Handle overflow
	    handleOverflow();
	}

        destructor {
		zapQueue();
	  }
}
