static const char file_id[] = "DECAPAlgorithm.pl";
// .cc file generated from DECAPAlgorithm.pl by ptlang
/*
 * copyright (c) SEE $PTOLEMY_SIMULATIONS/copyright
 */

#ifdef __GNUG__
#pragma implementation
#endif

#include "DECAPAlgorithm.h"

const char *star_nm_DECAPAlgorithm = "DECAPAlgorithm";

const char* DECAPAlgorithm :: className() const {return star_nm_DECAPAlgorithm;}

ISA_FUNC(DECAPAlgorithm,DEStar);

Block* DECAPAlgorithm :: makeNew() const { LOG_NEW; return new DECAPAlgorithm;}

DECAPAlgorithm::DECAPAlgorithm ()
{
	setDescriptor("This star simulates the switch behavior of the CAPC algorithm\n	       as was specified by Barnhart in ATM-Forum94-0983");
	addPort(IN_RM.setPort("IN_RM",this,MESSAGE));
	addPort(RET_RM.setPort("RET_RM",this,MESSAGE));
	addPort(Congestion.setPort("Congestion",this,INT));
	addPort(OUT_RM.setPort("OUT_RM",this,MESSAGE));
	addPort(NEW_RM.setPort("NEW_RM",this,MESSAGE));
	addState(PCR.setState("PCR",this,"0","Peak Cell Rate"));
	addState(target_rate.setState("target_rate",this,"0","target_rate"));
	addState(Rup.setState("Rup",this,"0.05","Initial value for the slope"));
	addState(ERU.setState("ERU",this,"1.5","maximum increase allowed"));
	addState(Rdn.setState("Rdn",this,"0.5","slope parameter"));
	addState(ERF.setState("ERF",this,"0.5","slope parameter"));
	addState(DQT.setState("DQT",this,"200","High Queue Limit"));
	addState(max_count.setState("max_count",this,"100","counting interval"));
	addState(interval.setState("interval",this,"1","timing interval"));
	addState(mode.setState("mode",this,"0","0=Explicit,1=Binary"));
	addState(RM_Proc.setState("RM_Proc",this,"","RM_Proc"));
	addState(debug.setState("debug",this,"0","debug"));
	addState(fileName.setState("fileName",this,"<stdout>","Filename for output"));

# line 157 "DECAPAlgorithm.pl"
p_out = 0;
}

void DECAPAlgorithm::wrapup() {
# line 160 "DECAPAlgorithm.pl"
LOG_DEL; delete p_out;
		p_out = 0;
}

void DECAPAlgorithm::begin() {
# line 148 "DECAPAlgorithm.pl"
LOG_DEL; delete p_out;
		LOG_NEW; p_out = new pt_ofstream(fileName);
}

DECAPAlgorithm::~DECAPAlgorithm() {
# line 163 "DECAPAlgorithm.pl"
LOG_DEL; delete p_out;
}

void DECAPAlgorithm::setup() {
# line 138 "DECAPAlgorithm.pl"
congested=0;
		count=0;
		start_time=0;
		ERS=target_rate/(float)10;
}

void DECAPAlgorithm::go() {
# line 166 "DECAPAlgorithm.pl"
pt_ofstream& OUTPUT = *p_out;
                completionTime= arrivalTime;

		InDEMPHIter inIter(RET_RM);
		InDEPort* inPtr;
		if(Congestion.dataNew)
		{
			Congestion.dataNew=FALSE;
			float cong=Congestion%0;
			congested= (int) cong;
		} //end Congestion

		if(IN_RM.dataNew)
	        {
			Envelope Envp;
	                IN_RM.get().getMessage(Envp);
       	            	ATMCell* cellPtr=(ATMCell *)
                    	  	Envp.myData();
                  	if(cellPtr->readRM()==0)
			{
				if((++count>=max_count)||
					(completionTime>start_time+interval))
				{
					float rate=(float)count/(float)
					   (completionTime-(float)start_time);
					float delta=(float)1-((float)rate/
							 (float)target_rate);
					float ERX;

					if(delta>=0)
					{
						ERX=(float)1+((float)delta*
							      (float)Rup);
						ERX=min((float)ERX,(float)ERU);
					}
					else
					{
						ERX=(float)1+((float)delta*
							      (float)Rdn);
						ERX=max((float)ERX,(float)ERF);
					}
					ERS=ERS*ERX;
					count=0;
					start_time=completionTime;
				}
			}
			OUT_RM.put(completionTime)<<Envp;
		}

//******************************************************************
		while ((inPtr = inIter++) != 0) 
		{
		     while (inPtr->dataNew) 
		     {
			Envelope Envp;
			inPtr->get().getMessage(Envp);
       	            	ATMCell* cellPtr=(ATMCell *)
                    	  	Envp.myData();
			int VP=cellPtr->readDestination();
			int VC=cellPtr->readVC();
		       if((VP!=RM_Proc[0])&&(VP!=RM_Proc[1])&&(VP!=RM_Proc[2])
			  &&(VP!=RM_Proc[3]))
		       {	           
                            Envelope envp( *cellPtr);
			    NEW_RM.put(completionTime)<<envp;
		            return;
			}
			float ACR=cellPtr->readACR();
                        LOG_NEW;
                        ATMCell* RM_Cell = new ATMCell(0);
                        RM_Cell->setRM(1);
                        RM_Cell->setSource(cellPtr->readSource());
                        RM_Cell->setVC(cellPtr->readVC());
                        RM_Cell->setDestination(cellPtr->readDestination());
                        RM_Cell->setDIR(0);
                        RM_Cell->setACR(ACR);
                        RM_Cell->setCI(cellPtr->readCI());
                        RM_Cell->setER(cellPtr->readER());
                        RM_Cell->setCI(cellPtr->readCI());
if(debug)
OUTPUT<<" cong "<<congested<<" ER "<<cellPtr->readER()<<" ERS "<<ERS<<" at "<<completionTime<<"\n";
			if(cellPtr->readER()>ERS)
				RM_Cell->setER(ERS);
			if(congested>DQT)
				RM_Cell->setCI(1);
if(debug)
OUTPUT<<" VP "<<VP<<" VC "<<VC<<" ER "<<RM_Cell->readER()<<" CI "<<RM_Cell->readCI()<<" at "<<completionTime<<"\n";
                        Envelope envp( *RM_Cell);
			NEW_RM.put(completionTime)<<envp;
		    }
		}// end RET_RM
}

// prototype instance for known block list
static DECAPAlgorithm proto;
static RegisterBlock registerBlock(proto,"CAPAlgorithm");
