static const char file_id[] = "DEOSU_Algorithm.pl";
// .cc file generated from DEOSU_Algorithm.pl by ptlang
/*
 * copyright (c) SEE $PTOLEMY_SIMULATIONS/copyright
 */

#ifdef __GNUG__
#pragma implementation
#endif

#include "DEOSU_Algorithm.h"

const char *star_nm_DEOSU_Algorithm = "DEOSU_Algorithm";

const char* DEOSU_Algorithm :: className() const {return star_nm_DEOSU_Algorithm;}

ISA_FUNC(DEOSU_Algorithm,DEStar);

Block* DEOSU_Algorithm :: makeNew() const { LOG_NEW; return new DEOSU_Algorithm;}

DEOSU_Algorithm::DEOSU_Algorithm ()
{
	setDescriptor("ABR Switch algorithm specified in ATM-Forum 94-883");
	addPort(IN_RM.setPort("IN_RM",this,MESSAGE));
	addPort(RET_RM.setPort("RET_RM",this,MESSAGE));
	addPort(OUT_RM.setPort("OUT_RM",this,MESSAGE));
	addPort(NEW_RM.setPort("NEW_RM",this,MESSAGE));
	addState(Number_VC.setState("Number_VC",this,"9","Number of possible VCs"));
	addState(Measurement_Int.setState("Measurement_Int",this,"1","Length of the measurement interval in cells"));
	addState(Target_Rate.setState("Target_Rate",this,"1","target rate in cells per sec"));
	addState(RM_Proc.setState("RM_Proc",this,"","RM_Proc"));
	addState(mode.setState("mode",this,"0","0=Explicit rate Notification 1=Intelligent Marking"));
	addState(fileName.setState("fileName",this,"<stdout>","Filename for output"));
	addState(debug.setState("debug",this,"0","debug"));

# line 119 "DEOSU_Algorithm.pl"
p_out = 0;
}

void DEOSU_Algorithm::wrapup() {
# line 122 "DEOSU_Algorithm.pl"
LOG_DEL; delete p_out;
		p_out = 0;
}

void DEOSU_Algorithm::begin() {
# line 110 "DEOSU_Algorithm.pl"
LOG_DEL; delete p_out;
		LOG_NEW; p_out = new pt_ofstream(fileName);
}

DEOSU_Algorithm::~DEOSU_Algorithm() {
# line 125 "DEOSU_Algorithm.pl"
LOG_DEL; delete p_out;
}

void DEOSU_Algorithm::setup() {
# line 93 "DEOSU_Algorithm.pl"
Cell_Count=0;
		start_interval=0;
		for(int i=0;i<12;i++)
		{
			VC_info[i].Cell_Seen=0;
			VC_info[i].Rev_RM_Seen=0;
			VC_info[i].Last_ER=0;
			VC_info[i].Last_ACR=0;
		}
		over_load=New_ER=0;
		fair_share=0;
}

void DEOSU_Algorithm::go() {
# line 128 "DEOSU_Algorithm.pl"
pt_ofstream& OUTPUT = *p_out;
                completionTime= arrivalTime;
		InDEMPHIter inIter(RET_RM);
		InDEPort* inPtr;

//**************************************************************************

		if(IN_RM.dataNew)
	        {
			Envelope Envp;
	                IN_RM.get().getMessage(Envp);
       	            	ATMCell* cellPtr=(ATMCell *)
                    	  	Envp.myData();
			int Destination=cellPtr->readDestination();
			int VC =cellPtr->readVC();
			int index;
			int VPs=(int)Number_VC/3;
			for(int l=0;l<VPs;l++)
		        {
		            if(Destination==RM_Proc[l])
			           index=l;
	                }
		        index=(index*3)+(VC%3);
			
                  	if(cellPtr->readRM()==0)
			{
				Cell_Count++;
			        OUT_RM.put(completionTime)<<Envp;
                                VC_info[index].Cell_Seen=1;
				if(Measurement_Int<=Cell_Count)
				{
				  float Interval_length=completionTime-
					start_interval;
				  start_interval=completionTime;
				  (float)over_load=(float)Cell_Count/
			 	    ((float)Target_Rate*(float)Interval_length);
if(debug) OUTPUT<<" count "<<Cell_Count<<" Interval "<<Interval_length;
			 	  int count=0;
				  for(int i=0;i<Number_VC;i++)
				  {
					if(VC_info[i].Cell_Seen)
						count++;
					VC_info[i].Cell_Seen=0;
				   }
				   fair_share=Target_Rate/(float)count;
if(debug) OUTPUT<<" OVERLOAD "<<over_load<<" fairshare "<<fair_share<<" count "<<count<<" at "<<completionTime<<"\n";
				   Cell_Count=0;
				   for(i=0;i<Number_VC;i++)
					VC_info[i].Rev_RM_Seen=0;

				   
				}
				return;
			}
			VC_info[index].Last_ACR=cellPtr->readACR();
			OUT_RM.put(completionTime)<<Envp;
if(debug)
OUTPUT<<" RECEIVER RM with ACR="<<VC_info[index].Last_ACR<<" RM  "<<cellPtr->readRM()<<"\n";
		}
//**************************************************************************
		while ((inPtr = inIter++) != 0) 
		{
		     while (inPtr->dataNew) 
		     {
			Envelope Envp;
			inPtr->get().getMessage(Envp);
       	            	ATMCell* cellPtr=(ATMCell *)
                    	  	Envp.myData();
			int VP=cellPtr->readDestination();
			int VC=cellPtr->readVC();
			int index;
		       if((VP!=RM_Proc[0])&&(VP!=RM_Proc[1])&&(VP!=RM_Proc[2])
                                  &&(VP!=RM_Proc[3]))
		       {	           
                            Envelope envp( *cellPtr);
			    NEW_RM.put(completionTime)<<envp;
		            return;
			}
			int VPs=(int)Number_VC/3;
			for(int l=0;l<VPs;l++)
		        {
		            if(VP==RM_Proc[l])
			           index=l;
	                }
                        index=((int)index*(int)3)+((int)VC%(int)3);

if(debug)
OUTPUT<<" index "<<index<<" VC "<<VC<<" Dest "<<VP<<" REV_RM "<<VC_info[index].Rev_RM_Seen<<"\n";
			if(VC_info[index].Rev_RM_Seen==0)
			{
				float ER1=(float)VC_info[index].Last_ACR/(float)over_load;
if(debug)
OUTPUT<<" index "<<index<<" Last_ACR "<<VC_info[index].Last_ACR<<" over_load "<<over_load<<" ER1 "<<ER1<<" at "<<completionTime<<"\n";
				float ER2=max(ER1,fair_share);
				New_ER=min((float)Target_Rate,(float)ER2);
					
				VC_info[index].Rev_RM_Seen=1;
				VC_info[index].Last_ER=New_ER;
			}
			else
			{
				New_ER=VC_info[index].Last_ER;
			}
                        LOG_NEW;
                        ATMCell* RM_Cell = new ATMCell(0);
                        RM_Cell->setRM(1);
                        RM_Cell->setSource(cellPtr->readSource());
                        RM_Cell->setVC(cellPtr->readVC());
                        RM_Cell->setDestination(cellPtr->readDestination());
                        RM_Cell->setDIR(0);
                        RM_Cell->setACR(cellPtr->readACR());
                        RM_Cell->setCI(cellPtr->readCI());
if(debug) OUTPUT<<"New_ER "<<New_ER<<" CER "<<cellPtr->readER()<<" LAST_ACR "<<VC_info[index].Last_ACR<<"\n";
	                if(mode)
		        {
                           RM_Cell->setER(cellPtr->readER());
		           if(New_ER<VC_info[index].Last_ACR)
		              RM_Cell->setCI(1);
			   else
		              RM_Cell->setCI(cellPtr->readCI());
			}
		        else
		        {
			   if(New_ER<cellPtr->readER())
			      RM_Cell->setER(New_ER);
			   else
                              RM_Cell->setER(cellPtr->readER());
			}
		        
if(debug)
OUTPUT<<" DEST "<<RM_Cell->readDestination()<<" ER "<<RM_Cell->readER()<<"\n";
                        Envelope envp( *RM_Cell);
			NEW_RM.put(completionTime)<<envp;

		    }
		}
}

// prototype instance for known block list
static DEOSU_Algorithm proto;
static RegisterBlock registerBlock(proto,"OSU_Algorithm");
