defstar {
 	name {OSU_Algorithm }
	domain { DE }
	author { Dorgham Sisalem}
        copyright { SEE $PTOLEMY_SIMULATIONS/copyright }

	desc {ABR Switch algorithm specified in ATM-Forum 94-883
	}
	input { name { IN_RM}
                type { message}
	        desc { incoming RM Cells}
	}

	inmulti { name { RET_RM}
                type { message}
	        desc { Returning RM Cells}
	}


	output { name { OUT_RM}
                 type { message}
	         desc { Send RM Cells to the Destination}
        }

	output { name { NEW_RM}
                 type { message}
	         desc { updated RM Cells}
	}
        defstate {
                name {Number_VC }
                type { int}
                default { 9}
                desc {Number of possible VCs}
        }

	defstate {
            
		name {Measurement_Int }
	        type { int}
		default { 1}
		desc {Length of the measurement interval in cells }
	}
	defstate {
		name {Target_Rate }
		type { float}
		default { 1}
		desc {target rate in cells per sec  }
	}
        defstate {
                name { RM_Proc}
                type { intarray}
        }
        defstate {
                name { mode }
                type { int }
                default { 0 }
	        desc{ 0=Explicit rate Notification 1=Intelligent Marking}
        }


	hinclude { "./VC.h","ATMCell.h" ,<math.h>,<minmax.h>}

	defstate {
		name { fileName }
		type { string }
		default { "<stdout>" }
		desc { Filename for output }
	}
        defstate {
                name { debug }
                type { int }
                default { 0 }
        }

	hinclude { "pt_fstream.h" }
	protected {
		pt_ofstream *p_out;
	}

	protected { VC_Data VC_info[12];
		    int Cell_Count;
		    float start_interval;
		    float fair_share,New_ER;
		    float over_load;
		    
		  }

//*****************************************************************************/
// initialization: 
//*****************************************************************************/

	setup { 
		Cell_Count=0;
		start_interval=0;
		for(int i=0;i<12;i++)
		{
			VC_info[i].Cell_Seen=0;
			VC_info[i].Rev_RM_Seen=0;
			VC_info[i].Last_ER=0;
			VC_info[i].Last_ACR=0;
		}
		over_load=New_ER=0;
		fair_share=0;
              }

//*****************************************************************************/
//*****************************************************************************/

	begin  {  
                LOG_DEL; delete p_out;
		LOG_NEW; p_out = new pt_ofstream(fileName);
		}

//****************************************************************************/
// set scheduling priorities
//****************************************************************************/

	constructor { 
                      p_out = 0;
               	    }
        wrapup {
	        LOG_DEL; delete p_out;
		p_out = 0;
		}
	destructor { LOG_DEL; delete p_out;}

	go {
                pt_ofstream& OUTPUT = *p_out;
                completionTime= arrivalTime;
		InDEMPHIter inIter(RET_RM);
		InDEPort* inPtr;

//**************************************************************************

		if(IN_RM.dataNew)
	        {
			Envelope Envp;
	                IN_RM.get().getMessage(Envp);
       	            	ATMCell* cellPtr=(ATMCell *)
                    	  	Envp.myData();
			int Destination=cellPtr->readDestination();
			int VC =cellPtr->readVC();
			int index;
			int VPs=(int)Number_VC/3;
			for(int l=0;l<VPs;l++)
		        {
		            if(Destination==RM_Proc[l])
			           index=l;
	                }
		        index=(index*3)+(VC%3);
			
                  	if(cellPtr->readRM()==0)
			{
				Cell_Count++;
			        OUT_RM.put(completionTime)<<Envp;
                                VC_info[index].Cell_Seen=1;
				if(Measurement_Int<=Cell_Count)
				{
				  float Interval_length=completionTime-
					start_interval;
				  start_interval=completionTime;
				  (float)over_load=(float)Cell_Count/
			 	    ((float)Target_Rate*(float)Interval_length);
if(debug) OUTPUT<<" count "<<Cell_Count<<" Interval "<<Interval_length;
			 	  int count=0;
				  for(int i=0;i<Number_VC;i++)
				  {
					if(VC_info[i].Cell_Seen)
						count++;
					VC_info[i].Cell_Seen=0;
				   }
				   fair_share=Target_Rate/(float)count;
if(debug) OUTPUT<<" OVERLOAD "<<over_load<<" fairshare "<<fair_share<<" count "<<count<<" at "<<completionTime<<"\n";
				   Cell_Count=0;
				   for(i=0;i<Number_VC;i++)
					VC_info[i].Rev_RM_Seen=0;

				   
				}
				return;
			}
			VC_info[index].Last_ACR=cellPtr->readACR();
			OUT_RM.put(completionTime)<<Envp;
if(debug)
OUTPUT<<" RECEIVER RM with ACR="<<VC_info[index].Last_ACR<<" RM  "<<cellPtr->readRM()<<"\n";
		}
//**************************************************************************
		while ((inPtr = inIter++) != 0) 
		{
		     while (inPtr->dataNew) 
		     {
			Envelope Envp;
			inPtr->get().getMessage(Envp);
       	            	ATMCell* cellPtr=(ATMCell *)
                    	  	Envp.myData();
			int VP=cellPtr->readDestination();
			int VC=cellPtr->readVC();
			int index;
		       if((VP!=RM_Proc[0])&&(VP!=RM_Proc[1])&&(VP!=RM_Proc[2])
                                  &&(VP!=RM_Proc[3]))
		       {	           
                            Envelope envp( *cellPtr);
			    NEW_RM.put(completionTime)<<envp;
		            return;
			}
			int VPs=(int)Number_VC/3;
			for(int l=0;l<VPs;l++)
		        {
		            if(VP==RM_Proc[l])
			           index=l;
	                }
                        index=((int)index*(int)3)+((int)VC%(int)3);

if(debug)
OUTPUT<<" index "<<index<<" VC "<<VC<<" Dest "<<VP<<" REV_RM "<<VC_info[index].Rev_RM_Seen<<"\n";
			if(VC_info[index].Rev_RM_Seen==0)
			{
				float ER1=(float)VC_info[index].Last_ACR/(float)over_load;
if(debug)
OUTPUT<<" index "<<index<<" Last_ACR "<<VC_info[index].Last_ACR<<" over_load "<<over_load<<" ER1 "<<ER1<<" at "<<completionTime<<"\n";
				float ER2=max(ER1,fair_share);
				New_ER=min((float)Target_Rate,(float)ER2);
					
				VC_info[index].Rev_RM_Seen=1;
				VC_info[index].Last_ER=New_ER;
			}
			else
			{
				New_ER=VC_info[index].Last_ER;
			}
                        LOG_NEW;
                        ATMCell* RM_Cell = new ATMCell(0);
                        RM_Cell->setRM(1);
                        RM_Cell->setSource(cellPtr->readSource());
                        RM_Cell->setVC(cellPtr->readVC());
                        RM_Cell->setDestination(cellPtr->readDestination());
                        RM_Cell->setDIR(0);
                        RM_Cell->setACR(cellPtr->readACR());
                        RM_Cell->setCI(cellPtr->readCI());
if(debug) OUTPUT<<"New_ER "<<New_ER<<" CER "<<cellPtr->readER()<<" LAST_ACR "<<VC_info[index].Last_ACR<<"\n";
	                if(mode)
		        {
                           RM_Cell->setER(cellPtr->readER());
		           if(New_ER<VC_info[index].Last_ACR)
		              RM_Cell->setCI(1);
			   else
		              RM_Cell->setCI(cellPtr->readCI());
			}
		        else
		        {
			   if(New_ER<cellPtr->readER())
			      RM_Cell->setER(New_ER);
			   else
                              RM_Cell->setER(cellPtr->readER());
			}
		        
if(debug)
OUTPUT<<" DEST "<<RM_Cell->readDestination()<<" ER "<<RM_Cell->readER()<<"\n";
                        Envelope envp( *RM_Cell);
			NEW_RM.put(completionTime)<<envp;

		    }
		}
	   } // end go
} // end defstar
